#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <tcp.h>
#include <string.h>
#include "vncdos.h"
#include "rfbproto.h"
#include "video.h"

/* vnc part */
CARD16 fb_width = 0;
CARD16 fb_height = 0;

#define BUF_IN_MAX 512
byte buf_in[BUF_IN_MAX];

int vncproto_init(tcp_Socket *fd, char *host, int port)
{
	longword addr;
	int status,i,s;

#define ENCODINGS 4
	rfbProtocolVersionMsg vmsg;
	rfbClientInitMsg clientinit;
	rfbServerInitMsg serverinit;
	rfbSetPixelFormatMsg pixformmsg;
	rfbSetEncodingsMsg *encodingsmsgp;
	CARD32 *enc;
	CARD32 i32;
	CARD8 x;
	size_t sz_enc;

#if 0
	CARD8 passwd[128];
#endif

	sz_enc=sz_rfbSetEncodingsMsg+ENCODINGS*sizeof(CARD32);
	encodingsmsgp = (rfbSetEncodingsMsg *)malloc(sz_enc);
	if (encodingsmsgp == NULL) {
		perror("malloc: Cannot initiate communication");
		return -1;
	}
	enc=(CARD32*)(sz_rfbSetEncodingsMsg+(CARD8*)encodingsmsgp);

	addr=resolve(host);
	if (addr==0L) {
		fprintf(stderr, "Cannot resolve hostname\n");
		return 0;
	}

	if (!tcp_open(fd, 0, addr, port, NULL)) {
		fprintf(stderr,"Cannot create socket\n");
		return 0;
	}

	sock_wait_established(fd, sock_delay, NULL, &status);
	
	sprintf(vmsg, rfbProtocolVersionFormat, rfbProtocolMajorVersion, rfbProtocolMinorVersion);
	sock_write(fd, vmsg, sz_rfbProtocolVersionMsg);
	sock_read(fd, vmsg, sz_rfbProtocolVersionMsg);

	i32 = rfbConnFailed;
	sock_read(fd, (byte*)&i32, sizeof(i32));
	i32 = ntohl(i32);

	switch(i32) {
		case rfbConnFailed: /* conn failed */
			puts("Remote server says: Connection failed");
			i32 = 0;
			sock_read(fd, (byte*)&i32, sizeof(i32));
			while (i32-- && (sock_read(fd, &x, sizeof(x))==sizeof(x)))
				printf ("%c", x);
			puts("");
			sock_close(fd);
			return 0;
		case rfbNoAuth:
			break;
		case rfbVncAuth:
			puts ("We don't support DES yet");
			sock_close(fd);
			return 0;
#if 0
			p = getpass("Enter password: ");
			if (!p) {
				sock_close(fd);
				return 0;
			}
#endif
	}

	/* ClientInitialisation */
	clientinit.shared = 1; // share
	sock_write(fd, (byte*)&clientinit, sz_rfbClientInitMsg);

	sock_read(fd, (byte*)&serverinit, sz_rfbServerInitMsg);

	fb_width = ntohs(serverinit.framebufferWidth);
	if (fb_width > DOS_XMAX)
		fb_width = DOS_XMAX;
	fb_height = ntohs(serverinit.framebufferHeight);
	if (fb_height > DOS_YMAX)
		fb_height = DOS_YMAX;

	i32 = ntohl(serverinit.nameLength);
	for (i=0; i<i32; i++)
		sock_read(fd, (byte*)&x, 1);
	
	pixformmsg.type = rfbSetPixelFormat;
	pixformmsg.format.bitsPerPixel = 8;
	pixformmsg.format.depth = 8;
	pixformmsg.format.bigEndian = 0; // don't care
	pixformmsg.format.trueColour = 1;

	pixformmsg.format.redMax = htons(3);
	pixformmsg.format.greenMax = htons(7);
	pixformmsg.format.blueMax = htons(7);

	pixformmsg.format.redShift = 0;
	pixformmsg.format.greenShift = 2;
	pixformmsg.format.blueShift = 5;

	sock_write(fd, (byte*)&pixformmsg, sz_rfbSetPixelFormatMsg);

	encodingsmsgp->type = rfbSetEncodings;
	encodingsmsgp->nEncodings = htons(ENCODINGS);
	enc[0] = htonl(rfbEncodingRaw);
	enc[1] = htonl(rfbEncodingCopyRect);
	enc[2] = htonl(rfbEncodingRRE);
	enc[3] = htonl(rfbEncodingCoRRE);
	sock_write(fd, (byte*)encodingsmsgp, sz_enc);

sock_err:
	switch (status) {
		case 1:
			fprintf(stderr,"remote host closed connection\n");
			return 0;
		case -1:
			fprintf(stderr,"server timeout: %s\n",sockerr(fd));
			return 0;
	}
	return 1;
}

int request_vnc_refresh(tcp_Socket *fd)
{
	rfbFramebufferUpdateRequestMsg updreq;
	static int incremental = 0;

	updreq.type = rfbFramebufferUpdateRequest;
	updreq.incremental = incremental;
	incremental=1;
	updreq.x = htons(0);
	updreq.y = htons(0);
	updreq.w = htons(fb_width);
	updreq.h = htons(fb_height);

	sock_write(fd, (byte*)&updreq, sz_rfbFramebufferUpdateRequestMsg);
	return 0;
}

int update_rect0(tcp_Socket *fd, rfbFramebufferUpdateRectHeader *uprect)
{
	CARD32	total;
	CARD32	pos;
	int	to_read, n_read;

	for (pos=0; pos<total;) {
		if (total-pos > BUF_IN_MAX) to_read = BUF_IN_MAX;
		else to_read=total-pos;
		
		n_read = sock_read(fd, buf_in, to_read);
		if (n_read == -1) {
			dbug_printf("fail at %ld\n",pos);
			return 0;
		}
		video_blk(uprect->r.x,uprect->r.y,
			  uprect->r.w,uprect->r.h,
			  pos, n_read, buf_in);
		pos += n_read;
	}
	return 1;
}

int parse_vnc_in(tcp_Socket *fd)
{
#define ST_IDLE		0
#define ST_CUTTXT	1
#define ST_RECT		2
#define ST_RAW		3
#define ST_COPY		4
#define ST_RRE		5
#define ST_CRRE		7
	static int state=ST_IDLE;
	static rfbFramebufferUpdateRectHeader uprect;
	static rfbRREHeader rrehead;
	CARD8 pixel;
	rfbRectangle subRect;
	rfbCoRRERectangle coRect;
	static rfbServerToClientMsg msg;
	static int rect;
	static CARD32 pos,total;
	rfbCopyRect copyrect;
	int i, s;

	if (state==ST_IDLE) {
		i = sock_read (fd, (byte*)&msg, sizeof(CARD8));
		if (i != sizeof(CARD8)) return 1;
	
		switch (msg.type) {
		case rfbFramebufferUpdate:
			i = sock_read(fd, ((byte*)&msg)+sizeof(CARD8), sz_rfbFramebufferUpdateMsg - sizeof(CARD8));
			msg.fu.nRects = ntohs(msg.fu.nRects);
			rect=0;
			state=ST_RECT;
			dbug_printf("msg: fbu: %d rectangles\n",msg.fu.nRects);
			break;

		case rfbBell:
			i = sock_read(fd, ((byte*)&msg)+sizeof(CARD8), sz_rfbBellMsg - sizeof(CARD8));
			fprintf(stderr,"\007\n");
			dbug_printf("msg: bell\n");
			break;

		case rfbServerCutText:
			i = sock_read(fd, ((byte*)&msg)+sizeof(CARD8), sz_rfbServerCutTextMsg - sizeof(CARD8));
			state=ST_CUTTXT;
			total=msg.sct.length;
			dbug_printf("msg: srv.cuttext: %ld\n",total);
			pos=0;
			break;

		default:
			dbug_printf("msg: unknown: %d\n",msg.type);
			return 0;
		}
	}

	if (state==ST_CUTTXT) {
		if (total-pos > BUF_IN_MAX) s = BUF_IN_MAX;
		else s=total-pos;
		
		i = sock_read(fd, buf_in, s);
		if (i == -1) {
			dbug_printf("fail server cut text at %ld\n",pos);
			return 0;
		}
		if (pos>=total) state=ST_IDLE;
		dbug_printf("  cuttext: %ld of %ld\n",pos,total);
	}

	if (state==ST_RECT) {
		if (rect >= msg.fu.nRects) state=ST_IDLE;
		else {
			i = sock_read(fd, (byte*)&uprect, sizeof(uprect));
			if (i != sizeof(uprect)) {
				state=ST_IDLE;
				return 0;
			}
			uprect.r.x = ntohs(uprect.r.x);
			uprect.r.y = ntohs(uprect.r.y);
			uprect.r.w = ntohs(uprect.r.w);
			uprect.r.h = ntohs(uprect.r.h);
			uprect.encoding = ntohl(uprect.encoding);
			dbug_printf("  rect: #%d (%d,%d) %dx%d, enc:%ld\n",
				rect, uprect.r.x, uprect.r.y,
				uprect.r.w, uprect.r.h,	uprect.encoding);
			
			if (uprect.r.x >= fb_width) 		 { state=ST_IDLE; return 0; }
			if (uprect.r.x+uprect.r.w > fb_width)	 { state=ST_IDLE; return 0; }
			if (uprect.r.y >= fb_height)		 { state=ST_IDLE; return 0; }
			if (uprect.r.y+uprect.r.h > fb_height)	 { state=ST_IDLE; return 0; }

			switch(uprect.encoding) {
			case rfbEncodingRaw:
				total=(CARD32)uprect.r.w * (CARD32)uprect.r.h;
				pos=0;
				state=ST_RAW;
				break;
			case rfbEncodingCopyRect:
				state=ST_COPY;
				break;
			case rfbEncodingRRE:
			case rfbEncodingCoRRE:
				i = sock_read(fd, (byte*)&rrehead,sz_rfbRREHeader);
				if (i != sz_rfbRREHeader) { state=ST_IDLE; return 0; }
				total=ntohl(rrehead.nSubrects);
				i = sock_read(fd, (byte*)&pixel,sizeof(CARD8));
				if (i != sizeof(CARD8)) { state=ST_IDLE; return 0; }
				dbug_printf("    enc_rre: %ld subrectangles\n",
					total);
				
				drawbar(uprect.r.x,uprect.r.y,
					uprect.r.w,uprect.r.h, video_pixcolor(&pixel));

				pos=0;
				state=ST_RRE;
				if (uprect.encoding==rfbEncodingCoRRE)
				state=ST_CRRE;
				break;
			}
		}
	}

	if (state==ST_RAW) {
		if (total-pos > BUF_IN_MAX) s = BUF_IN_MAX;
		else s=total-pos;
		
		i = sock_read(fd, buf_in, s);
		if (i == -1) {
			dbug_printf("fail at %ld\n",pos);
			return 0;
		}
		video_blk(uprect.r.x,uprect.r.y,
			  uprect.r.w,uprect.r.h,
			  pos, i, buf_in);
		pos += i;
		if (pos>=total) {
			rect++;
			state=ST_RECT;
		}
		dbug_printf("    enc_raw: %ld of %ld bytes\n",pos,total);
	}

	if (state==ST_COPY) {
		i = sock_read(fd, (byte*)&copyrect, sz_rfbCopyRect);
		if (i == -1) {
			dbug_printf("fail at copyrect\n");
			return 0;
		}
		copyrect.srcX = ntohs(copyrect.srcX);
		copyrect.srcY = ntohs(copyrect.srcY);
		video_blt(uprect.r.x,uprect.r.y,
			  uprect.r.w,uprect.r.h,
			  copyrect.srcX,copyrect.srcY);
		rect++;
		state=ST_RECT;
		dbug_printf("    enc_copytrct: from (%d,%d)\n",pos,total);
	}

	if (state==ST_RRE) {
		i = sock_read(fd, (byte*)&pixel,sizeof(CARD8));
		if (i != sizeof(CARD8)) { state=ST_IDLE; return 0; }
		i = sock_read(fd, (byte*)&subRect,sz_rfbRectangle);
		if (i != sz_rfbRectangle) { state=ST_IDLE; return 0; }
		subRect.x=ntohs(subRect.x);
		subRect.y=ntohs(subRect.y);
		subRect.w=ntohs(subRect.w);
		subRect.h=ntohs(subRect.h);
		drawbar(uprect.r.x+subRect.x,uprect.r.y+subRect.y,
			subRect.w,subRect.h, video_pixcolor(&pixel));
		pos++;
		if (pos>=total) {
			rect++;
			state=ST_RECT;
		}
	}

	if (state==ST_CRRE) {
		i = sock_read(fd, (byte*)&pixel,sizeof(CARD8));
		if (i != sizeof(CARD8)) { state=ST_IDLE; return 0; }
		i = sock_read(fd, (byte*)&coRect,sz_rfbCoRRERectangle);
		if (i != sz_rfbCoRRERectangle) { state=ST_IDLE; return 0; }
		drawbar(uprect.r.x+coRect.x,uprect.r.y+coRect.y,
			coRect.w,coRect.h, video_pixcolor(&pixel));
		pos++;
		if (pos>=total) {
			rect++;
			state=ST_RECT;
		}
	}

	return 1;
}

int parse_kbd_in(int kbd, tcp_Socket *fd)
{
	static rfbKeyEventMsg ke = { rfbKeyEvent, 0, 0, 0};
	static rfbPointerEventMsg me = { rfbPointerEvent, 0, 0, 0};
	static int mouse_on = -1;
	static int mouse_factor = 1;
	static int ctrllock_state = 0;
	static int shiftlock_state = 0;
	CARD16 k;

	if (mouse_on == -1) {
		me.x = htons(DOS_XMAX/2);
		me.y = htons(DOS_YMAX/2);
		mouse_on = 0;
	}

	k = 0;

	switch(kbd) {
		case 0x129:  // ALT-ESC - exit
			return 0;
		case '\x08': k = 0xFF08; break;
		case '\x09': k = 0xFF09; break;
		case '\x1b': k = 0xFF1b; break;
		case 0x14b: k = 0xFF51; break;
		case 0x148: k = 0xFF52; break;
		case 0x14d: k = 0xFF53; break;
		case 0x150: k = 0xFF54; break;
		case 0x152: k = 0xFF63; break;
		case 0x153: k = 0xFFFF; break;
		case 0x147: k = 0xFF50; break;
		case 0x14f: k = 0xFF57; break;
		case 0x149: k = 0xFF55; break;
		case 0x151: k = 0xFF56; break;
//		case 'P': mouse_on = !mouse_on; break;
//		case 'Q': ctrllock_state ^= 1;
//			ke.down = ctrllock_state;
//			ke.key = 0xFFE3;
//			sock_write(fd, (byte*)&ke, sizeof(ke));
//			break;
//		case 'R': 
//			shiftlock_state ^= 1;
//			ke.down = shiftlock_state;
//			ke.key = 0xFFE1;
//			sock_write(fd, (byte*)&ke, sizeof(ke));
//			break;
		case '\x0d': 
			if (!mouse_on) {
				k = 0xFF0D; break;
			}
		case '1': case '2': case '3':
		case '4': case '5': case '6':
		case '7': case '8': case '9':
		case '0': case '.': case '-':
		case ',':
			if (!mouse_on) {
				k = kbd;
				break;
			}
			// DPRINT("mouse key\n");
			if (kbd == '5') {
				mouse_factor <<= 1;
				if (mouse_factor > 64)
					mouse_factor = 1;
				break;
			}
			if (kbd == '4' || kbd == '7' || kbd == '1')
				if (ntohs(me.x)>mouse_factor)
					me.x = htons(ntohs(me.x)-mouse_factor);
				else
					me.x = htons(0);
			if (kbd == '6' || kbd == '9' || kbd == '3')
				if (ntohs(me.x)+mouse_factor < DOS_XMAX-1)
					me.x = htons(ntohs(me.x)+mouse_factor);
				else
					me.x = htons(DOS_XMAX-1);
			if (kbd == '7' || kbd == '8' || kbd == '9')
				if (ntohs(me.y)>mouse_factor)
					me.y = htons(ntohs(me.y)-mouse_factor);
				else
					me.y = htons(0);
			if (kbd == '1' || kbd == '2' || kbd == '3')
				if (ntohs(me.y)+mouse_factor < DOS_YMAX-1)
					me.y = htons(ntohs(me.y)+mouse_factor);
				else
					me.y = htons(DOS_YMAX-1);
			if (kbd>='1' && kbd<='9') {
				sock_write(fd, (byte*)&me, sizeof(me));
				break;
			}
			if (kbd=='-') {
				me.buttonMask = me.buttonMask ^ rfbButton1Mask;
				sock_write(fd, (byte*)&me, sizeof(me));
			}
			if (kbd==',') {
				me.buttonMask = me.buttonMask ^ rfbButton2Mask;
				sock_write(fd, (byte*)&me, sizeof(me));
			}
			if (kbd=='\x0d') {
				me.buttonMask = me.buttonMask ^ rfbButton3Mask;
				sock_write(fd, (byte*)&me, sizeof(me));
			}
			if (kbd=='0') {
				me.buttonMask = me.buttonMask ^ rfbButton1Mask;
				sock_write(fd, (byte*)&me, sizeof(me));
				me.buttonMask = me.buttonMask ^ rfbButton1Mask;
				sock_write(fd, (byte*)&me, sizeof(me));
			}
			if (kbd=='.') {
				me.buttonMask = me.buttonMask ^ rfbButton2Mask;
				sock_write(fd, (byte*)&me, sizeof(me));
				me.buttonMask = me.buttonMask ^ rfbButton2Mask;
				sock_write(fd, (byte*)&me, sizeof(me));
			}
			break;
		default: if (kbd < 0x100) k = kbd;
				else k=0;
			break;
	}

	if (k>0) {
		ke.down = 1;
		ke.key = htonl(k);
		sock_write(fd, (byte*)&ke, sizeof(ke));
		ke.down = 0;
		sock_write(fd, (byte*)&ke, sizeof(ke));
	}
	
	return 1;
}
